/***********************************************************************/
/*                                                                     */
/*                                OCaml                                */
/*                                                                     */
/*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
/*                                                                     */
/*  Copyright 1996 Institut National de Recherche en Informatique et   */
/*  en Automatique.  All rights reserved.  This file is distributed    */
/*  under the terms of the Q Public License version 1.0.               */
/*                                                                     */
/***********************************************************************/

/* Simplified OCaml syntax */

%{
open Location
open Asttypes
open Parsetree

let mkpat d =
  { ppat_desc = d; ppat_loc = symbol_rloc() }
let mkexp d =
  { pexp_desc = d; pexp_loc = symbol_rloc() }
let mkrhs rhs pos = mkloc rhs (rhs_loc pos)
let mkoperator name pos =
  let loc = rhs_loc pos in
  { pexp_desc = Pexp_ident(mkloc (name) loc); pexp_loc = loc }

let reloc_pat x = { x with ppat_loc = symbol_rloc () };;
let reloc_exp x = { x with pexp_loc = symbol_rloc () };;

let ghexp d = { pexp_desc = d; pexp_loc = symbol_gloc () };;
let ghpat d = { ppat_desc = d; ppat_loc = symbol_gloc () };;

let mkinfix arg1 name arg2 =
  mkexp(Pexp_apply(arg2, mkexp(Pexp_apply(arg1, mkoperator name 2))))

let mkuminus arg =
  match arg.pexp_desc with
  | Pexp_constant(Const_int n) ->
      mkexp(Pexp_constant(Const_int(-n)))
  | _ ->
      mkexp(Pexp_apply(mkoperator "-u" 1, arg))

let syntax_error () =
  raise Syntaxerr.Escape_error

let unclosed opening_name opening_num closing_name closing_num =
  raise(Syntaxerr.Error(Syntaxerr.Unclosed(rhs_loc opening_num, opening_name,
                                           rhs_loc closing_num, closing_name)))

%}

/* Tokens */

%token AMPERAMPER
%token AMPERSAND
%token BARBAR
%token BEGIN
%token END
%token EOF
%token EQUAL
%token FUN
%token GREATER
%token IN
%token <int> INT
%token LESS
%token LET
%token <string> LIDENT
%token LPAREN
%token MINUS
%token MINUSGREATER
%token OR
%token PLUS
%token REC
%token RPAREN
%token SEMISEMI
%token STAR
%token <string> UIDENT
%token UNDERSCORE

/* Precedences and associativities */

%nonassoc IN
%nonassoc below_SEMI
%nonassoc LET                           /* above SEMI ( ...; let ... in ...) */
%right    MINUSGREATER                  /* core_type2 (t -> t -> t) */
%right    OR BARBAR                     /* expr (e || e || e) */
%right    AMPERSAND AMPERAMPER          /* expr (e && e && e) */
%nonassoc below_EQUAL
%left     EQUAL LESS GREATER            /* expr (e OP e OP e) */
%left     PLUS MINUS                    /* expr (e OP e OP e) */
%left     STAR                          /* expr (e OP e OP e) */
%nonassoc prec_unary_minus              /* unary - */
%nonassoc below_SHARP
/* Finally, the first tokens of simple_expr are above everything else. */
%nonassoc BEGIN INT LIDENT LPAREN UIDENT

%start implementation                   /* for implementation files */
%type <Parsetree.expression list> implementation
%start toplevel                         /* for interactive use */
%type <Parsetree.expression> toplevel
%%

/* Entry points */

implementation:
    structure EOF                        { $1 }
;
toplevel:
  | seq_expr SEMISEMI                    { $1 }
  | EOF                                  { raise End_of_file }
;
structure:
    structure_tail                              { $1 }
  | seq_expr structure_tail                     { $1 :: $2 }
;
structure_tail:
    /* empty */                                 { [] }
  | SEMISEMI                                    { [] }
  | SEMISEMI seq_expr structure_tail            { $2 :: $3 }
;

/* Patterns */

pattern:
    simple_pattern
      { $1 }
;
simple_pattern:
    val_ident %prec below_EQUAL
      { mkpat($1) }
  | LPAREN pattern RPAREN
      { reloc_pat $2 }
  | LPAREN pattern error
      { unclosed "(" 1 ")" 3 }
;

/* Core expressions */

seq_expr:
  | expr   %prec below_SEMI  { $1 }
;
expr:
    simple_expr_list %prec below_SHARP
      { $1 }
  | LET rec_flag pattern EQUAL seq_expr IN seq_expr
      { mkexp(Pexp_let($2, $3, $5, $7)) }
  | FUN simple_pattern fun_def
      { mkexp(Pexp_function ($2, $3)) }
  | expr PLUS expr
      { mkinfix $1 "+" $3 }
  | expr MINUS expr
      { mkinfix $1 "-" $3 }
  | expr STAR expr
      { mkinfix $1 "*" $3 }
  | expr EQUAL expr
      { mkinfix $1 "=" $3 }
  | expr LESS expr
      { mkinfix $1 "<" $3 }
  | expr GREATER expr
      { mkinfix $1 ">" $3 }
  | expr OR expr
      { mkinfix $1 "or" $3 }
  | expr BARBAR expr
      { mkinfix $1 "||" $3 }
  | expr AMPERSAND expr
      { mkinfix $1 "&" $3 }
  | expr AMPERAMPER expr
      { mkinfix $1 "&&" $3 }
  | MINUS expr %prec prec_unary_minus
      { mkuminus $2 }
;
simple_expr:
    val_ident
      { mkexp(Pexp_ident (mkrhs ($1) 1)) }
  | INT
      { mkexp(Pexp_constant (Const_int $1)) }
  | LPAREN seq_expr RPAREN
      { reloc_exp $2 }
  | LPAREN seq_expr error
      { unclosed "(" 1 ")" 3 }
  | BEGIN seq_expr END
      { reloc_exp $2 }
  | BEGIN seq_expr error
      { unclosed "begin" 1 "end" 3 }
;
simple_expr_list:
    simple_expr %prec below_SHARP
      { $1 }
  | simple_expr_list simple_expr
      { mkexp(Pexp_apply($2, $1)) }
;
fun_binding:
    EQUAL seq_expr
      { $2 }
  | simple_pattern fun_binding
      { ghexp(Pexp_function($1, $2)) }
;
match_action:
    MINUSGREATER seq_expr                       { $2 }
;
fun_def:
    match_action                                { $1 }
  | simple_pattern fun_def
      { ghexp(Pexp_function($1, $2)) }
;

/* Identifiers */
val_ident:
    LIDENT                                      { $1 }
  | LPAREN operator RPAREN                      { $2 }
;
operator:
    PLUS                                        { "+" }
  | MINUS                                       { "-" }
  | STAR                                        { "*" }
  | EQUAL                                       { "=" }
  | LESS                                        { "<" }
  | GREATER                                     { ">" }
  | OR                                          { "or" }
  | BARBAR                                      { "||" }
  | AMPERSAND                                   { "&" }
  | AMPERAMPER                                  { "&&" }
;

/* Miscellaneous */

rec_flag:
    /* empty */                                 { Nonrecursive }
  | REC                                         { Recursive }
;
